#include <WiFi.h>
#include "ESPAsyncWebServer.h"
#include "SPIFFS.h"
#include <ArduinoJson.h>
#include "Arduino_JSON.h"
#include <math.h>

// ESP32のADCライブラリをインクルード
#include "driver/adc.h" 
#include "esp_adc_cal.h"
esp_adc_cal_characteristics_t adcChar;

// WiFi-AP関連
const char ssid[] = "ESP32AP_WEB_SERVICE";
const char pass[] = "31415926";       // パスワードは8文字以上
const IPAddress ip(192,168,21,23);
const IPAddress subnet(255,255,255,0);
AsyncWebServer server(80);            // ポート設定

// Pin assignment
#define LED_PIN_GP25 25
#define THERMISTOR_PIN 33
#define PWM_DRV 23
#define PWM_DRV_G 22
#define PWM_FAN 15

#define ANALOG_MAX 3300     // 3300 mv
#define ARRAY_SIZE 5
#define U_MAX 256.0
#define U_MIN 0.0
#define MODE_MED 0
#define MODE_AVE 1


// グローバル変数の定義と初期化
//温調制御関係
String gMethod = "PID";
String order = "STOP";
String previous_order = "STOP";
int gTestCommand = 0;
// PID control parapeter //
float gKp = 40.0;
float gKi = 0.01;
float gKd = 5.0;
float gTime = 0.05;       // unit is "second"
float gTarget = 41.0;
float gPre_Error = 0.0;
float gI = 0.0;
float gM_T = 0.0;
float gM_P =0;
float gM_I =0;
float gM_D =0;

// Thermistor(103AT-11)用を用いたサーミスタ回路からの電圧読み取り用の定数
const float R1 = 10000.00; // 分圧抵抗10kΩ
const float TK = 273.15;
const float R0 = 27280.00;
const float B = 3435.00;

//　ON/OFF制御用変数
int gPrevious_u = 0;
float gTh_higher = 0.00;
float gTh_lower = 0.00;
float gHys = 0.15;

// 温調の操作量＝DRV8871へ渡すPWM信号（0～256）
int gM_duty = 0;
int gM_U = 0;

// 異常値排除（5回電圧値を読み取り昇順に並べ変えて3番目値を採用ことでノイズを採用しないようにする）
uint32_t  getVoltage(int MODE){
  uint32_t voltage;
  uint32_t voltage_array[ARRAY_SIZE];
  uint32_t voltageESP;
  uint32_t temp;
  int i,j,p,q;
  temp = 0.00;
  i = j = p = q = 0;


  for(i=0; i < ARRAY_SIZE; i++){
    esp_adc_cal_get_voltage(ADC_CHANNEL_5, &adcChar, &voltageESP);     //ADC1_CH5 (=GPIO33)の電圧値を読み取り
	  voltage_array[i] = voltageESP;
  }
  
  if (MODE == MODE_AVE){  // average value
    // sorting
    for (p = 0; p < (ARRAY_SIZE - 1); p++) {
      for (q = (ARRAY_SIZE - 1); q > p; q--) {
        if (voltage_array[q-1] > voltage_array[q]) {
          temp = voltage_array[q-1];
          voltage_array[q-1] = voltage_array[q];
          voltage_array[q] = temp;
        }
      }
    }
    // 突発値を除いた上で平均値を得る
    voltage = (voltage_array[1]+voltage_array[2]+voltage_array[3]) / 3.00;

  } else {   // median value
    // sorting
    for (p = 0; p < (ARRAY_SIZE - 1); p++) {
      for (q = (ARRAY_SIZE - 1); q > p; q--) {
        if (voltage_array[q-1] > voltage_array[q]) {
          temp = voltage_array[q-1];
          voltage_array[q-1] = voltage_array[q];
          voltage_array[q] = temp;
        }
      }
    }
    voltage = voltage_array[2];
  }
    
  
  return voltage;
}

float getTemp(){
  float Aout, temp, T_bar, Rt;
  uint32_t voltage;
  
  voltage = getVoltage(MODE_AVE);
  Aout =  (float)voltage;
  Rt = (ANALOG_MAX/Aout - 1) * R1;
  T_bar = 1/TK + 1/B * log(Rt/R0);
  temp = 1.0/T_bar - TK;

  return temp;
}

void init_gPre_Error(float target){
  gPre_Error = target - getTemp();
}

// PID control function
int getControlInput_PID(){
  float y, r, e, d, u;
  
  y = getTemp();                                  //  現在の温度を取得
  r = gTarget;                                    //  目標温度を取得
  e = r - y;                                      //  目標温度と現在温度の差を計算
  //if ( e < 0) e = 0;
  
  d   = ( e - gPre_Error) / gTime;                //  誤差の微分の近似計算
  gI  = gI + ( e + gPre_Error ) * (gTime/2);      //  誤差の積分の近似計算
  u   = ( gKp * e ) + ( gKi * gI ) + ( gKd * d);  //  操作量を計算
  // for debug monitor
  gM_T =  y;
  gM_P =  gKp * e ;
  gM_I =  gKi * gI;
  gM_D =  gKd * d;
  // for debug END
  
  if(u > U_MAX) u = U_MAX;
  if(u < U_MIN) u = U_MIN;
  gPre_Error = e;                                 //  次回の計算に備えて今回の誤差を保存

  
  return (int)(u+0.5);
}

// ON-OFF control function
int getControlInput_ONOFF(){
  float currentTemperature;
  int u;
  
  currentTemperature = getTemp();
  if (currentTemperature < gTh_lower) {
    // 現在の温度が目標温度-ヒステリシスより低い場合、ヒーターをONにする
    u = 256;
  } else if (currentTemperature > gTh_higher) {
    // 現在の温度が目標温度+ヒステリシスより高い場合、ヒーターをOFFにする
    u = 0;
  } else {
    // 他の条件の場合は、前回の操作量／命令を維持する
    u = gPrevious_u;
  }
  gPrevious_u = u;
  return u;
}

void setup()
{
  Serial.begin(115200);
 
  adc_gpio_init(ADC_UNIT_1, ADC_CHANNEL_5);                                                 // ADC1_CH5を初期化 (=GPIO33)
  adc1_config_width(ADC_WIDTH_BIT_12);                                                      // ADC1の解像度を12bit（0~4095）に設定
  adc1_config_channel_atten(ADC1_CHANNEL_6, ADC_ATTEN_DB_11);                               // ADC1の減衰を11dBに設定
  esp_adc_cal_characterize(ADC_UNIT_1, ADC_ATTEN_DB_11, ADC_WIDTH_BIT_12, 1100, &adcChar);  // 電圧値に変換するための情報をaddCharに格納

  pinMode(LED_PIN_GP25, OUTPUT);   //  GPIO25を出力設定に
  digitalWrite(LED_PIN_GP25, LOW);
  pinMode(PWM_DRV_G, OUTPUT);       
  digitalWrite(PWM_DRV_G, LOW);    //  GPIO22を出力設定に(DRVモーターは正転のみ使用するのでIN2を0v固定にしておく)
  ledcSetup(0, 1000,8);            //  チャンネル0, キャリア周波数1kHz、8ビットレンジ
  ledcAttachPin(PWM_DRV, 0);       //  PWM DRVピンにチャンネル0を指定
  ledcSetup(1, 25000,8);           //  チャンネル1, キャリア周波数25kHz、8ビットレンジ
  ledcAttachPin(PWM_FAN, 1);       //  PWM FANピンにチャンネル0を指定

    
  // SPIFFSのセットアップ
  if(!SPIFFS.begin(true)){
    Serial.println("An Error has occurred while mounting SPIFFS");
    return;
  }
  
  // WiFi-APの設定設定
  WiFi.softAP(ssid, pass);           // SSIDとパスの設定
  delay(100);                        // このdelayを入れないと失敗する場合がある
  WiFi.softAPConfig(ip, ip, subnet); // IPアドレス、ゲートウェイ、サブネットマスクの設定
  IPAddress myIP = WiFi.softAPIP();  // WiFi.softAPIP()でWiFi起動

  // 各種情報を表示
  Serial.print("SSID: ");
  Serial.println(ssid);
  Serial.print("AP IP address: ");
  Serial.println(myIP);

  // GETリクエストに対するハンドラーを登録
  // rootにアクセスされた時のレスポンス
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/index.html");
  });
  // style.cssにアクセスされた時のレスポンス
  server.on("/style.css", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/style.css", "text/css");
  });
  // script.jsにアクセスされた時のレスポンス
  server.on("/script.js", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/script.js");
  });

  //デバイスへの命令に対する処理
   server.on(
     "/post_ctrl",
     HTTP_POST,
     [](AsyncWebServerRequest * request){},
     NULL,
     [](AsyncWebServerRequest * request, uint8_t *data, size_t len, size_t index, size_t total) {
       String ctrl_json;
       for (size_t i = 0; i < len; i++) {
         //Serial.write(data[i]);
         ctrl_json.concat(char(data[i]));
       }
       // Jsonオブジェクトの初期化  using ArduinoJson.h
       StaticJsonDocument<512> command; 
       Serial.println(ctrl_json);
       DeserializationError error = deserializeJson(command, ctrl_json);

       if(error){
         Serial.println("deserializeJson() faild");
         request->send(400);
       }
       else{
        
        //JSONデータから命令やデータを抽出する
        order = (const char*)command["START/STOP"];
        gMethod = (const char*)command["ctrl"];
        gTarget = command["target"];

        // PID制御用パラメータの取得
        gKp = command["kp"];
        gKi = command["ki"];
        gKd = command["kd"];
        
        // ON-OFF制御用パラメータの取得
        gTh_higher = gTarget + gHys;
        gTh_lower = gTarget - gHys;

        // 命令実行の判断
        if( (previous_order == "STOP") && ( order == "START") ) {
          Serial.println("Temperature control Started!"); 
          gTestCommand = 1;
          init_gPre_Error(gTarget);
          gI = 0.0;

        } else if ((previous_order == "START") && (order == "STOP") ){
          //Serial.println("Temperature control Stopped!");
          gTestCommand = 0;
        } else {

        }
        // 現在の命令を保存
        previous_order = order;

         // 受け取ったJSONデータをループバックさせる
         request->send(200, "application/json", ctrl_json);
       }
      //  送信されてくるJSONデータ："START/STOP":"STOP", "target":33.0,"kp":29.0, "ki":1.5, "kd":5.2, "ctrl":"PID"
   });


  // デバイスへのデータ取得リクエストに対する処理
  server.on(
    "/post_sensor",
    HTTP_POST,
    [](AsyncWebServerRequest * request){

      JSONVar sensor_json; // using Arduino_JSON.h

      gM_T = getTemp();
      sensor_json["TEMP"] = gM_T;
      sensor_json["U"] = gM_U; // 
      sensor_json["P"] = gM_P;
      sensor_json["I"] = gM_I;
      sensor_json["D"] = gM_D;
      sensor_json["E"] = gPre_Error;
      String result = JSON.stringify(sensor_json);
      request->send(200, "application/json", result);
      //Serial.println("Get /post_sensor");

  });

  // サーバースタート
  server.begin();
  Serial.println("Server start.....");
}


void loop() {
  
  // testCommand状態変更
  if(gTestCommand == 0){
    digitalWrite(LED_PIN_GP25, LOW);
    gM_duty = 0;
  }
  else{
    digitalWrite(LED_PIN_GP25, HIGH);
    if (gMethod == "PID"){
      gM_U = getControlInput_PID();
    } else {
      gM_U = getControlInput_ONOFF();
    }
    
    gM_duty = gM_U;
  }
  ledcWrite(0, gM_duty);    // PWM DRV
  
  if ( gM_T >= 35.0) {
    ledcWrite(1, 256);      // PWM FAN DUTY:100%
  } else if ( gM_T >= 33.0){
    ledcWrite(1, 192);      // PWM FAN DUTY:75%
  } else {
    ledcWrite(1, 128);      // PWM FAN DUTY:50%
  }
     
  
  delay(gTime*1000);
}
